/******************************************************************************
Watershed filter plugin for Avisynth by V.C.Mohan
Watershed algorithm after vincent- soille. Useful for segmentation of image 
through water shed mapping.  Immersion technique is used.
Individual catchment basins can be displayed 
Author V.C.Mohan
July 18, 2005, April 6,2009 Aug 7, 2014
Copyright (C) <2005-- 2014>  <V.C.Mohan>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    A copy of the GNU General Public License is at
    http://www.gnu.org/licenses/.
	For details of how to contact author see <http://www.avisynth.org/vcmohan/vcmohan.html> 
********************************************************************************/  
// use x64 platform
#define AVISYNTHPLUS
// use win32 platform
//#define AVISYNTH_VERSION_2_6

#ifdef AVISYNTHPLUS
#include "windows.h"
#ifndef AVISYNTH_VERSION_2_6
#define AVISYNTH_VERSION_2_6
#endif
//#define AVS_LINKAGE_DLLIMPORT  // This is for static linkage, so not required
#include "avisynthPlus.h"
const AVS_Linkage *AVS_linkage = 0;	// for dynamic linkage
#else
#ifdef AVISYNTH_VERSION_2_6
#include "windows.h"
#include "avisynth_2_6.h"
const AVS_Linkage *AVS_linkage = 0;	// for dynamic linkage

#else
#include "windows.h"
#include "avisynth.h"	// 2.5.8 version

#endif
#endif
#include <algorithm>
//#include <queue>
//#include <stack>		
#include <functional>
#include "WSSegment.cpp"

		



class Watershed : public GenericVideoFilter {
		
		int nl; //start frame number of  (child) clip
		int enl; // end frame number for this line
		bool connect4;		// true for 4 connect, false for 8 connect
		bool white;		// true image is white and background black, false for vice versa
		 
	
		int slx;			// start x coordinate of rectangle left 
		int sty;			// start y coordinate of rectangle top  
		int srx;			// start x coord of window right 
		int sby;			// start y coord of window bottom 
		int elx;			// end x coordinate of rectangle left 
		int ety;			// end y coordinate of rectangle top  
		int erx;			// end x coord of window right 
		int eby;			// end y coord	of window bottom 
		int slt,sut,flt,fut;	// start and end lower and upper thresholds
		bool test;			// is this a test run?
		bool grey;			// grey output window?
		bool showbasins;	// show basins?
		int slb, sub, elb, eub;	// %age lower and upper basin number at start and at end to show
		bool watershed;			// true (default) for display of watershed lines
		const unsigned char **pixelsort1;		// sort buffer
		int *tag1;			// basin label  buffer
		int *dist1;			// distance marked buffer
		char *ws1;			// watershed line mark buffer

		unsigned char  *buf1;	// buffer to hold values temporarily

		bool exist;			// flag for creation by new
		int numCPU;
		int psize;
  public:
							//Definition of function
    Watershed(PClip _child,int _nl,int _enl,
				bool _connect4, bool _white,
				
				int _slx,int _sty,int _srx, int _sby,
				int _elx,int _ety,int _erx, int _eby,
				int _slt,int _sut, int _flt, int _fut,
				bool _test, bool _grey, bool _showbasins,
				int _slb, int _sub, int _elb, int _eub,
				bool _watershed,
				IScriptEnvironment* env) ;	
			
	~Watershed();				//destructor

	PVideoFrame __stdcall GetFrame(int n, IScriptEnvironment* env);
//	void Erode_Dilator(const unsigned char *lp, unsigned char *wp,int spitch, int wpitch, int ww, int ht,int bytes, int seek);

// This is the function that AviSynth calls to get a given frame.
// So when this functions gets called, the filter is supposed to return frame n.

// Facilitates same calls of IsY8(), GetPlaneWidthSubsampling(int pl)
// and GetPlaneHeightSubsampling(int pl). 

#include "Planar_2_5_or_2_6.cpp"
};

class LessThan{
	public :
	bool operator()( const unsigned char *f1,const unsigned char *f2)
				{ return *f1 < *f2; }
};		
/*************************************************
 * The following is the implementation 
 * of the defined functions.
 *************************************************/
//Here is the acutal constructor code used

Watershed::Watershed(PClip _child,int _nl,int _enl,
				bool _connect4, bool _white,
				
				int _slx,int _sty,int _srx, int _sby,
				int _elx,int _ety,int _erx, int _eby,
				int _slt,int _sut, int _flt, int _fut,
				bool _test,bool _grey,bool _showbasins,
				int _slb, int _sub, int _elb, int _eub,
				bool _watershed,
				IScriptEnvironment* env) :

	GenericVideoFilter(_child),nl(_nl),enl(_enl),
						connect4(_connect4), white(_white), 
						
						slx(_slx), sty(_sty),
						srx(_srx), sby(_sby),
						elx(_elx), ety(_ety),
						erx(_erx), eby(_eby),
						slt(_slt), sut(_sut),
						flt(_flt), fut(_fut),
						test(_test), grey(_grey),
						showbasins(_showbasins),
						slb(_slb), sub(_sub), elb(_elb), eub(_eub),
						watershed(_watershed)						
						 
			
{
	char * Tname="Watershed";	
	
	psize = (srx - slx < erx - elx ? erx-elx +1 : srx - slx + 1) 
					* (sby - sty < eby - ety ? eby-ety +1 : sby - sty + 1); 

	numCPU = 1;	// safety for single processor or avisynth 2.6, 2.5.8

#ifdef AVISYNTHPLUS

	SYSTEM_INFO sysinfo;
	GetSystemInfo(&sysinfo);

	numCPU = sysinfo.dwNumberOfProcessors;
#endif

	if (numCPU == 1)
	{
		// no multi thread processing. So buffers can be created for common use		

		dist1 = new int[psize];

		ws1 = new char[psize];

		pixelsort1 = new const unsigned char*[psize];

		tag1 = new int[psize];

		buf1 = new unsigned char[psize];
	}
	
}

 Watershed::~Watershed()
{
	if (numCPU == 1)
	{
		delete[]dist1;
		delete[]ws1;
		delete[]pixelsort1;
		delete[]tag1;
		delete[]buf1;
	}
}

/***************************************************************/	
	PVideoFrame __stdcall Watershed::GetFrame(int in, IScriptEnvironment* env)
	{		
		 
		PVideoFrame Frame;
		int n, nframes;

		if(test)
		{					
			if(in <enl)
				Frame = child->GetFrame(nl, env);
			else
			{				
							
				return(child->GetFrame(in, env));
			}
			n=in;
			nframes=enl;
		}

		else
		{
			Frame = child->GetFrame(in, env);
			if(in<nl ||in>enl )
			{
			
				return Frame;
			}
	
			n=in-nl;
			nframes= enl-nl;	// actual intervals
			if ( nframes == 0) nframes = 1;
		}		
		// create or rename buffers
		int * dist = dist1;

		char * ws = ws1;

		const unsigned char ** pixelsort = pixelsort1;

		int * tag = tag1;

		unsigned char * buf = buf1;

#ifdef AVISYNTHPLUS
		// BEWARE BEWARE BEWARE
		// get buffers from pool. Iscript environment2 is not yet firm so may change aand break code.
		// IScriptEnvironment2* env2 = static_cast<IScriptEnvironment2*>(env);
		// use local names

		if (numCPU > 1)
		{
			
			IScriptEnvironment2* env2 = (IScriptEnvironment2*)(env);

			dist = (int *)(env2->Allocate(sizeof(int) * psize, 32, AVS_POOLED_ALLOC));
			tag = (int *)(env2->Allocate(sizeof(int) * psize, 32, AVS_POOLED_ALLOC));
			ws = (char*)(env2->Allocate(psize, 16, AVS_POOLED_ALLOC));
			buf = (unsigned char*)(env2->Allocate(psize, 16, AVS_POOLED_ALLOC));
			pixelsort = (const unsigned char * *)(env2->Allocate(sizeof(const unsigned char *) * psize, 64, AVS_POOLED_ALLOC));

		}
#endif	

		const unsigned char *fp= Frame->GetReadPtr();
		
		const	int bwd = vi.width;
        const	int bht = Frame->GetHeight();		
		const	int fpitch = Frame->GetPitch();
		int kb=vi.BitsPerPixel()/8;

		PVideoFrame work = env->NewVideoFrame(vi);

		unsigned char *wp= work->GetWritePtr();
		const	int wpitch = work->GetPitch();
		
		int object= white?255:0;
		int backg= white?0:255;

				// calculate parameters for this frame
		
		int lx=slx+((elx-slx)*n)/nframes;
		int ty=sty+((ety-sty)*n)/nframes;
		int by=sby+((eby-sby)*n)/nframes;
		int rx=srx+((erx-srx)*n)/nframes;
		int lt=slt+((flt-slt)*n)/nframes;
		int ut=sut+((fut-sut)*n)/nframes;
		int lb=slb+((elb-slb)*n)/nframes;
		int ub=sub+((eub-sub)*n)/nframes;
		

		if(vi.IsRGB())
		{
			// so that same code can be used for all color formats
			by=bht-1-(sty+((ety-sty)*n)/nframes);
			ty=bht-1-(sby+((eby-sby)*n)/nframes);
		}
		
		// current sizes of buffers needed
		int p1size = (rx - lx + 1) * (by - ty + 1);
			// we require a sort of input image grey values, but also have their original locations
						
			// we need to include  <algorithm> and <functional> for sort and <queue> for fifo queue


		int pitch = (rx-lx+1);	// similar to frames so that height and width can be used

		
		if(vi.IsYUY2() || vi.IsPlanar() || vi.IsRGB())
					
		{
			int hpitch = 0, hfpitch = ty * fpitch;
			for (int h = 0; h <= by - ty; h++)
			{
				for (int w = 0; w <= rx - lx; w++)
				{

					buf[hpitch + w] = fp[hfpitch + kb*(w + lx)];
				}
				hpitch += pitch;
				hfpitch += fpitch;
			}
			for(int h=0; h < p1size; h++)
						
					// sort buffer fill with pointers to data
					pixelsort[h] = buf + h ;
			
			// sort in ascending order the pointers pointing to
			std::sort(pixelsort, pixelsort+p1size, LessThan());

			WSSegment( pixelsort, buf, tag, dist, ws, by - ty + 1, rx - lx + 1, connect4);			

			int maxbasin =0;	
			if(showbasins)				 
					 // get max value of basin number for calculating %ages					 
			 for(int h = 0;h < p1size;h++)
										
				maxbasin = maxbasin >= tag[h]? maxbasin:tag[h];
			int lowerbasin = (maxbasin * lb) / 100;
			int upperbasin = (maxbasin * ub) / 100;

			double scale = 255.0 / (upperbasin-lowerbasin);	 

			env->BitBlt(wp, wpitch, fp, fpitch, work->GetRowSize(),bht);

			if(vi.IsPlanar() && !  IsY8())
			{
				const unsigned char * fpU = Frame->GetReadPtr(PLANAR_U);
				const unsigned char * fpV = Frame->GetReadPtr(PLANAR_V);
				const int  fpitchUV = Frame->GetPitch(PLANAR_U);
				const int bwdUV = Frame->GetRowSize(PLANAR_U);
				const int bhtUV = Frame->GetHeight(PLANAR_U);
				unsigned char * wpU = work->GetWritePtr(PLANAR_U);
				unsigned char * wpV = work->GetWritePtr(PLANAR_V);
				const int  wpitchUV = work->GetPitch(PLANAR_U);

				env->BitBlt(wpU, wpitchUV, fpU, fpitchUV, bwdUV,bhtUV);
				env->BitBlt(wpV, wpitchUV, fpV, fpitchUV, bwdUV,bhtUV);
			}
			
			int wshed=0;
				// mark water sheds 
			if(watershed && showbasins)
			{
				int hpitch = 0, hwpitch = ty * wpitch;

				for (int h = 0; h <= by - ty; h++)
				{
					for (int w = 0; w <= rx - lx; w++)
					{

						if (ws[hpitch + w] == wshed)

							wp[hwpitch + kb * (w + lx)] = object;

						else
						{
							if (tag[hpitch + w] >= lowerbasin && tag[hpitch + w] <= upperbasin)

								wp[hwpitch + kb * (w + lx)] = (tag[hpitch + w] - lowerbasin) * scale;
							else
								wp[hwpitch + kb * (w + lx)] = backg;
						}

					}
					hpitch += pitch;
					hwpitch += wpitch;
				}
			}

			else if(showbasins)
			{
				int hpitch = 0, hwpitch = ty * wpitch;
				for (int h = 0; h <= by - ty; h++)
				{
					for (int w = 0; w <= rx - lx; w++)
					{
						if (tag[hpitch + w] >= lowerbasin && tag[hpitch + w] <= upperbasin)

							wp[hwpitch + kb * (w + lx)] = (tag[hpitch + w] - lowerbasin) * scale;
						else
							wp[hwpitch + kb * (w + lx)] = backg;
					}
					hpitch += pitch;
					hwpitch += wpitch;
				}
			}

			else if(watershed)
			{
				int hpitch = 0, hwpitch = ty * wpitch;

				for (int h = 0; h <= by - ty; h++)
				{
					for (int w = 0; w <= rx - lx; w++)
					{
						if (ws[hpitch + w] == wshed)

							wp[hwpitch + kb * (w + lx)] = object;

						else
							wp[hwpitch + kb * (w + lx)] = backg;
					}
					hpitch += pitch;
					hwpitch += wpitch;
				}
			}
				 
		
				// grey border

			for(int h = ty;h < by;h++)
			{
				wp [h * wpitch + kb * lx] = 128;
				wp [h * wpitch + kb * rx] = 128;
			}

			for(int w = lx; w < rx; w++)
			{
				wp [ty * wpitch + kb * w] = 128;
				wp [by * wpitch + kb * w] = 128;
			}

			if(test & grey)
			{
				//  grey  u and v channels
				if(vi.IsYUY2())
				{
					for(int h = ty;h < by; h++)
						for(int w = lx;w < rx; w++)						
							wp[h * wpitch + kb * w + 1] = 128;
				}

				else if(vi.IsPlanar() && ! IsY8())
				{
				
					unsigned char * wpU = work->GetWritePtr(PLANAR_U);
					unsigned char * wpV = work->GetWritePtr(PLANAR_V);
					const int  wpitchUV = work->GetPitch(PLANAR_U);
					int	subW = (GetPlaneWidthSubsampling(PLANAR_U));	// bit shift number
					int	subH = (GetPlaneHeightSubsampling(PLANAR_U));

					for(int h = ty >> subH; h < by >> subH; h++)
						for(int w = lx >> subW; w < rx >> subW; w++)
						{
						// grey u and v channels
							wpU[h * wpitchUV + w] = 128;
							wpV[h * wpitchUV + w] = 128;
						}
				}

				else if(vi.IsRGB())
				{
					for(int h=ty;h<by;h++)
						for(int w=lx;w<rx;w++)
						{
								// grey by making  red green  blue equal .
							wp [h * wpitch + kb * w + 1] = wp[h * wpitch + kb * w + 2 ] = wp[h * fpitch + kb *w ];
						}
				}
			}


		}
#ifdef AVISYNTHPLUS
		if (numCPU > 1)
		{
			IScriptEnvironment2* env2 = (IScriptEnvironment2*)(env);
			env2->Free(buf);
			env2->Free(ws);
			env2->Free(tag);
			env2->Free(pixelsort);
			env2->Free(dist);

		}
#endif
							
			
return work;

		
	
}


/***************************************************************/
// This is the function that created the filter, when the filter has been called.
// This can be used for simple parameter checking, so it is possible to create different filters,

AVSValue __cdecl Create_Watershed(AVSValue args, void* user_data, IScriptEnvironment* env) 
{
	char * Tname = "Watershed";
	const VideoInfo& vi = args[0].AsClip()->GetVideoInfo();
#ifdef AVISYNTH_VERSION_2_6
	if( !vi.IsPlanar() &&  !vi.IsRGB()   && !vi.IsYUY2() )
		env->ThrowError("%s: This color format is not supported here", Tname);


#else
	if (!vi.IsRGB24() && !vi.IsRGB32()  && !vi.IsYV12() && !vi.IsYUY2())

		env->ThrowError("%s: This color format is not supported by this old avisynth version", Tname);
#endif		
		
	int nf=vi.num_frames;
	if(args[2].AsInt(0)<0 || args[2].AsInt(0)> args[3].AsInt(nf-1))
		env->ThrowError("Watershed: Frame number must be positive and end frame not less than start frame");
	if(args[2].AsInt(0)>=nf || args[3].AsInt(nf-1)>=nf)
		env->ThrowError("Watershed:this clip which has only %d frames",nf);

	
	int xx= vi.width-1;
	int yy= vi.height-1;
	if(xx>2047 || yy >2047)
		env->ThrowError("Watershed:frame x or y dimension should not exceed 2048");
	
	if(args[5].AsInt(0)<0 || args[6].AsInt(0) <0 || args[9].AsInt(0)<0 || args[10].AsInt(0)<0)
		env->ThrowError("Watershed: coordinates should not be -ve");
	if(args[7].AsInt(xx)<2 || args[8].AsInt(yy) <2 || args[11].AsInt(xx)<2 || args[12].AsInt(yy)<2)
		env->ThrowError("Watershed: dimensions should not be less than 2");
	if((args[7].AsInt(xx)>xx || args[8].AsInt(yy)>yy
		 || args[11].AsInt(args[7].AsInt(xx))>xx || args[12].AsInt(args[8].AsInt(yy))>yy) )
		env->ThrowError("Watershed:The dimensions pushes the window outside frame");
	if((args[5].AsInt(0)>args[7].AsInt(xx)-2 || args[6].AsInt(0)>args[8].AsInt(yy)-2
		 || args[9].AsInt(args[5].AsInt(0))>args[11].AsInt(args[7].AsInt(xx))-2 
		 || args[10].AsInt(args[6].AsInt(0))>args[12].AsInt(args[8].AsInt(yy))-2) )
		env->ThrowError("Watershed:The window has invalid coordinates. It must be atleast 2 pixel wide and high");

	if(args[13].AsInt(0) <0 || args[13].AsInt(0)> args[14].AsInt(255) || args[14].AsInt(255)>255)
		env->ThrowError("Watershed:lower and upper thresholds must be between 0 and 255, and lt must be less than ut");
	if(args[15].AsInt(args[13].AsInt(0)) <0 || args[15].AsInt(args[13].AsInt(0))>= args[16].AsInt(args[14].AsInt(255)) 
		|| args[16].AsInt(args[14].AsInt(255))>255)
		env->ThrowError("Watershed:final lower and upper thresholds must be between 0 and 255, and fut  more than flt");
	
		return  new Watershed(	args[0].AsClip(),			// clip on which areas are to be filled
								
								args[1].AsInt(0),			// start frame number to process	
								args[2].AsInt(nf-1),		// end frame number to process
								args[3].AsBool(true),		// true for 4 connect, false for 8 connect
								args[4].AsBool(true),		// true for image is white else black;
							
								args[5].AsInt(0),			// rectangle start left x, 
								args[6].AsInt(0),			// rectangle start top y,  
								args[7].AsInt(xx),			// start right x 
								args[8].AsInt(yy),			// start bottom y  
								args[9].AsInt(args[5].AsInt(0)),		// rectangle end left x, 
								args[10].AsInt(args[6].AsInt(0)),		// rectangle end top y,  
								args[11].AsInt(args[7].AsInt(xx)),		// end rightx 
								args[12].AsInt(args[8].AsInt(yy)),		// end bot y
								args[13].AsInt(0),			// (starting) lower threshold
								args[14].AsInt(255),		// (starting) upper threshold
								args[15].AsInt(args[13].AsInt(0)),			// ending lower threshold
								args[16].AsInt(args[14].AsInt(255)),		// ending upper threshold
								args[17].AsBool(false),		// is this a test?
								args[18].AsBool(false),		// grey out window?
								args[19].AsBool(true),		// show basins?
								args[20].AsInt(0),			// start lower %age threshold of basin numbers to show
								args[21].AsInt(100),			// start upper %age of upper threshold of basin numbers to show
								args[22].AsInt(args[20].AsInt(0)),			// ending lower %age thrshold of basin numbers to show
								args[23].AsInt(args[21].AsInt(100)),		// ending upper threshold of basin numbers to show
								args[24].AsBool(true),						// watershed lines to be shown?
								
								
								env);
	



	
}
// The following function is the function that actually registers the filter in AviSynth
// It is called automatically, when the plugin is loaded to see which functions this filter contains.

#ifdef AVISYNTH_VERSION_2_6
 
extern "C" __declspec(dllexport) const char* __stdcall AvisynthPluginInit3(IScriptEnvironment* env, const AVS_Linkage* const vectors)
{
	// Save the server pointers.
	AVS_linkage = vectors;
#else

extern "C" __declspec(dllexport) const char* __stdcall AvisynthPluginInit2(IScriptEnvironment* env)
{
#endif

    env->AddFunction("Watershed", "c[sf]i[ef]i[c4]b[white]b[lx]i[ty]i[rx]i[by]i[elx]i[ety]i[erx]i[eby]i[lt]i[ut]i[elt]i[eut]i[test]b[grey]b[showbasins]b[lb]i[ub]i[elb]i[eub]i[watershed]b", Create_Watershed, 0);
					
    return "Watershed  Plugin";
}

